from __future__ import annotations
from typing import TYPE_CHECKING, Callable, List, Optional, TypeVar, Union

from weakref import WeakSet
import warnings

if TYPE_CHECKING:  # pragma: no cover
    from .protocols import DisposableProtocol


_T = TypeVar("_T")


class Scope:
    def __init__(self, suite: ScopeSuite, detached=False) -> None:
        self._suite = suite
        self._active = True
        self._detached = detached
        self._disposables: WeakSet[DisposableProtocol] = WeakSet()
        self._cleanups: List[Callable[[], None]] = []
        self._parent: Optional[Scope] = None
        self._scopes: List[Scope] = []

        # child scope's index in parent's `self._scopes`
        self._index = -1

        self._parent = suite._ACTIVE_SCOPE
        if not detached and suite._ACTIVE_SCOPE:
            suite._ACTIVE_SCOPE._scopes.append(self)
            self._index = len(suite._ACTIVE_SCOPE._scopes) - 1

    @property
    def active(self):
        return self._active

    def run(self, fn: Callable[[], _T]) -> Union[_T, None]:
        if self.active:
            current_scope = self._suite._ACTIVE_SCOPE
            try:
                self._suite._ACTIVE_SCOPE = self
                return fn()
            finally:
                self._suite._ACTIVE_SCOPE = current_scope
        else:
            warnings.warn("cannot run inactive scope.")

    def on(self):
        """
        only be called on non-detached scopes
        """
        self._suite._ACTIVE_SCOPE = self

    def off(self):
        """
        only be called on non-detached scopes
        """
        self._suite._ACTIVE_SCOPE = self._parent

    def add_disposable(self, disposable: DisposableProtocol):
        self._disposables.add(disposable)

    def dispose(self, from_parent=False):
        if self._active:
            for effect in self._disposables:
                effect.dispose()

            for cleanup in self._cleanups:
                cleanup()

            for child_scope in self._scopes:
                child_scope.dispose(True)

            if (
                (not self._detached)
                and self._parent
                and (not from_parent)
                and self._parent._scopes
            ):
                last = self._parent._scopes.pop()
                if last is not self:
                    # swap position with the last element
                    self._parent._scopes[self._index] = last
                    last._index = self._index

            self._disposables.clear()
            self._parent = None
            self._active = False

    # def mark_with_scope(self, effect: DisposableProtocol):
    #     scope = scope or self._suite._ACTIVE_SCOPE
    #     if scope:
    #         scope._add_disposable(effect)


class ScopeSuite:
    def __init__(self) -> None:
        self._ACTIVE_SCOPE: Optional[Scope] = None

    def scope(self, detached=False):
        return Scope(self, detached)

    def mark_with_scope(self, effect: DisposableProtocol):
        if self._ACTIVE_SCOPE:
            self._ACTIVE_SCOPE.add_disposable(effect)

    def get_current_scope(
        self,
    ):
        return self._ACTIVE_SCOPE

    def on_scope_dispose(self, callback: Callable[[], None]):
        if self._ACTIVE_SCOPE:
            self._ACTIVE_SCOPE._cleanups.append(callback)
        else:
            warnings.warn("There are no active scopes.")


_DEFAULT_SCOPE_SUITE = ScopeSuite()

scope = _DEFAULT_SCOPE_SUITE.scope



from typing import Callable, Generic, TypeVar, cast

from signe.core.types import TMaybeSignal, TSignal

_T = TypeVar("_T")


class ReadableMixin(Generic[_T]):
    __slots__ = ("__weakref__",)

    @property
    def value(self) -> _T:
        ...


def is_signal(obj: TMaybeSignal):
    """Checks if a value is a signal or computed object.

    Args:
        obj (_type_): _description_
    """
    return isinstance(obj, ReadableMixin)


def to_value(obj: TMaybeSignal[_T]) -> _T:
    """Normalizes values / signals / getters to values.

    Args:
        obj (_type_): _description_

    ## Example
    ```
    to_value(1)          #    --> 1
    to_value(signal(1))  #    --> 1
    ```

    """
    if is_signal(obj):
        return cast(TSignal, obj).value

    if isinstance(obj, Callable):
        return obj()

    return cast(_T, obj)



from __future__ import annotations
from typing import Any, Dict, Optional, Set, TYPE_CHECKING
from signe.core.idGenerator import IdGen
from .consts import EffectState


if TYPE_CHECKING:  # pragma: no cover
    from .computed import Computed
    from signe.core.protocols import CallerProtocol
    from .runtime import ExecutionScheduler


class Dep:
    _id_gen = IdGen("Dep")

    def __init__(self, computed: Optional[Computed] = None) -> None:
        self.__id = Dep._id_gen.new()
        self.computed = computed
        self._deps: Set[CallerProtocol] = set()

    def get_callers(self):
        return tuple(self._deps)

    def add_caller(self, caller: CallerProtocol):
        self._deps.add(caller)

    def remove_caller(self, caller: CallerProtocol):
        self._deps.remove(caller)

    def __hash__(self) -> int:
        return hash(self.__id)

    def __eq__(self, __value: object) -> bool:  # pragma: no cover
        if isinstance(__value, self.__class__):
            return self.__hash__() == __value.__hash__()

        return False


class GetterDepManager:
    def __init__(
        self,
        scheduler: ExecutionScheduler,
    ) -> None:
        self._scheduler = scheduler
        self._deps_map: Dict[str, Dep] = {}

    def tracked(
        self, key, value: Optional[Any] = None, computed: Optional[Computed] = None
    ):
        running_caller = self._scheduler.get_running_caller()

        if not (running_caller and self._scheduler.should_track()):
            return

        dep = self._deps_map.get(key)
        if not dep:
            dep = Dep(computed)
            self._deps_map[key] = dep

        dep.add_caller(running_caller)
        running_caller.add_upstream_ref(dep)

    def triggered(self, key, value, state: EffectState):
        dep = self._deps_map.get(key)
        if not dep:
            return

        scheduler = self._scheduler

        for caller in dep.get_callers():
            caller.trigger(state)

        if scheduler.should_run:
            scheduler.run()

    def dispose(self):
        self._deps_map.clear()



from signe.core.signal import Signal, signal
from signe.core.mixins import to_value, is_signal
from signe.core.effect import Effect, effect, stop
from signe.core.computed import Computed, computed
from signe.core.asyncComputed import async_computed
from signe.core.runtime import ExecutionScheduler
from signe.core.batch import batch
from signe.core.on import on, WatchedState
from signe.core.cleanup import cleanup
from signe.core.reactive import reactive, to_raw, is_reactive
from signe.core.scope import scope
from signe.core.types import TMaybeSignal, TGetterSignal, TSignal, TGetter

__all__ = [
    "Signal",
    "Effect",
    "Computed",
    "async_computed",
    "ExecutionScheduler",
    "signal",
    "effect",
    "computed",
    "batch",
    "on",
    "to_value",
    "is_signal",
    "cleanup",
    "reactive",
    "scope",
    "stop",
    "to_raw",
    "TMaybeSignal",
    "TGetterSignal",
    "TGetter",
    "TSignal",
    "WatchedState",
    "is_reactive",
]

__version__ = "0.4.20"



from typing import Any, Callable, Coroutine, TypeVar, Union
from signe.core.protocols import ComputedResultProtocol, SignalResultProtocol


_T = TypeVar("_T")
_T_async_fn = Callable[[], Coroutine[Any, Any, _T]]

TSignal = SignalResultProtocol[_T]
TGetterSignal = Union[SignalResultProtocol[_T], ComputedResultProtocol[_T]]
TGetter = Union[TGetterSignal[_T], Callable[[], _T]]
TMaybeSignal = Union[_T, TGetter[_T]]



from __future__ import annotations
from typing import (
    Any,
    Dict,
    List,
    Callable,
    Optional,
    TypeVar,
    Union,
    cast,
    Generic,
    overload,
    TYPE_CHECKING,
)


from signe.core.idGenerator import IdGen
from signe.core.scope import Scope, ScopeSuite, _DEFAULT_SCOPE_SUITE

from .consts import EffectState
from .context import get_default_scheduler
from functools import partial

if TYPE_CHECKING:  # pragma: no cover
    from signe.core.deps import Dep
    from .runtime import ExecutionScheduler


_T = TypeVar("_T")


class EffectOption:  # pragma: no cover
    def __init__(self, level=0) -> None:
        self.level = level


class Effect(Generic[_T]):
    _id_gen = IdGen("Effect")

    def __init__(
        self,
        fn: Callable[[], _T],
        *,
        scheduler: ExecutionScheduler,
        scope: Union[Scope, ScopeSuite],
        trigger_fn: Optional[Callable[[Effect], None]] = None,
        scheduler_fn: Optional[Callable[[Effect], None]] = None,
        debug_trigger: Optional[Callable] = None,
        priority_level=1,
        debug_name: Optional[str] = None,
        capture_parent_effect=True,
        state: Optional[EffectState] = None,
    ) -> None:
        self.__id = self._id_gen.new()
        self._scheduler = scheduler
        self._active = True
        self._fn = fn
        self._trigger_fn = trigger_fn
        self._scheduler_fn = scheduler_fn
        self._upstream_refs: Dict[Dep, None] = {}
        self._debug_name = debug_name
        self._debug_trigger = debug_trigger

        self._state: EffectState = state or EffectState.NEED_UPDATE
        self._cleanups: List[Callable[[], None]] = []

        if isinstance(scope, Scope):
            scope.add_disposable(self)
        elif isinstance(scope, ScopeSuite):
            scope.mark_with_scope(self)

        self._sub_effects: List[Effect] = []

        if capture_parent_effect:
            running_caller = self._scheduler.get_running_caller()
            if running_caller and running_caller.is_effect:
                cast(Effect, running_caller).made_sub_effect(self)

    @property
    def id(self):
        return self.__id

    @property
    def state(self):
        return self._state

    @property
    def is_effect(self) -> bool:
        return True

    def calc_state(self):
        if self.state == EffectState.NEED_UPDATE:
            return

        self._state = EffectState.QUERYING
        # self._executor.pause_track()

        for dep in self._upstream_refs.keys():
            if dep.computed:
                dep.computed.confirm_state()
                if self._state <= EffectState.NEED_UPDATE:
                    break

        if self._state == EffectState.QUERYING:
            self._state = EffectState.STALE

        # self._executor.reset_track()

    def is_need_update(self):
        if not self._active:
            return False
        self.calc_state()
        return self.state <= EffectState.NEED_UPDATE

    def made_sub_effect(self, sub: Effect):
        self._sub_effects.append(sub)

    def trigger(self, state: EffectState):
        scheduler = self._scheduler
        scheduler.pause_scheduling()
        self._state = state

        if self._trigger_fn:
            self._trigger_fn(self)

        if self._scheduler_fn:
            self._scheduler.push_scheduler_fn(partial(self._scheduler_fn, self))

        scheduler.reset_scheduling()

    def add_upstream_ref(self, dep: Dep):
        self._upstream_refs[dep] = None

    def update_state(self, state: EffectState):
        self._state = state

    def update(self) -> _T:
        if not self._active:
            return self._fn()

        try:
            self._exec_cleanups()
            self._scheduler.mark_running_caller(self)
            self._state = EffectState.RUNNING

            self._clear_all_deps()

            self._dispose_sub_effects()
            result = self._fn()
            if self._debug_trigger:
                self._debug_trigger()

            return result

        finally:
            self._state = EffectState.STALE
            self._scheduler.reset_running_caller(self)

    def stop(self):
        self.dispose()

    def _clear_all_deps(self):
        for dep in self._upstream_refs:
            dep.remove_caller(self)

        self._upstream_refs.clear()

    def dispose(self):
        self._active = False
        self._clear_all_deps()
        self._exec_cleanups()
        self._dispose_sub_effects()

    def _dispose_sub_effects(self):
        for sub in self._sub_effects:
            sub.dispose()

        self._sub_effects.clear()

    def _exec_cleanups(self):
        for fn in self._cleanups:
            fn()

        self._cleanups.clear()

    def add_cleanup(self, fn: Callable[[], None]):
        self._cleanups.append(fn)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, __value: object) -> bool:  # pragma: no cover
        if isinstance(__value, self.__class__):
            return self.__hash__() == __value.__hash__()

        return False

    def __call__(self) -> Any:
        return self.update()

    def __repr__(self) -> str:
        return f"Effect(id ={self.id}, name={self._debug_name})"


_TEffect_Fn = Callable[[Callable[..., _T]], Effect]


@overload
def effect(
    fn: None = ...,
    *,
    immediate=True,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> _TEffect_Fn[None]:
    ...


@overload
def effect(
    fn: Callable[[], None],
    *,
    immediate=True,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> Effect:
    ...


def effect(
    fn: Optional[Callable[[], None]] = None,
    *,
    immediate=True,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> Union[_TEffect_Fn[None], Effect]:
    kws = {
        "priority_level": priority_level,
        "debug_trigger": debug_trigger,
        "debug_name": debug_name,
        "immediate": immediate,
        "scheduler": scheduler or get_default_scheduler(),
    }

    if fn:
        if isinstance(fn, Effect):
            fn = fn._fn

        scope = scope or _DEFAULT_SCOPE_SUITE
        # executor = get_executor()

        def scheduler_fn(effect: Effect):
            if effect.is_need_update():
                effect.update()

        kws.pop("immediate")
        res = Effect(fn, scheduler_fn=scheduler_fn, **kws, scope=scope)
        if immediate:
            res.update()
        return res
    else:

        def wrap(fn: Callable[..., None]):
            return effect(fn, **kws, scope=scope)

        return wrap


def stop(effect: Effect):
    effect.stop()



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Sequence,
    TypeVar,
    Callable,
    Optional,
    Union,
    cast,
)
from signe.core.mixins import ReadableMixin
from signe.core.signal import signal
from signe.core.context import get_default_scheduler
from signe.core.on import on

from signe.core.protocols import ComputedResultProtocol

from .types import TGetter, TSignal, _T_async_fn
from .scope import Scope, ScopeSuite, _DEFAULT_SCOPE_SUITE

if TYPE_CHECKING:  # pragma: no cover
    from .runtime import ExecutionScheduler

_T = TypeVar("_T")


_T_wrap_fn = Callable[[_T_async_fn], ComputedResultProtocol[_T]]


def async_computed(
    source: Union[TGetter, Sequence[TGetter]],
    *,
    init: Optional[_T] = None,
    evaluating: Optional[TSignal[bool]] = None,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> _T_wrap_fn[_T]:
    scheduler = scheduler or get_default_scheduler()

    def wrap_cp(fn: _T_async_fn):
        current = signal(init, is_shallow=True, scheduler=scheduler)
        evaluating_ref = evaluating or signal(False, is_shallow=True)
        evaluating_ref.value = False

        effect_kws = {
            "debug_name": debug_name,
            "debug_trigger": debug_trigger,
        }

        @on(
            source,
            onchanges=True,
            effect_kws=effect_kws,
            scheduler=scheduler,
            scope=scope or _DEFAULT_SCOPE_SUITE,
        )
        async def _():
            evaluating_ref.value = True

            try:
                current.value = await fn()
            finally:
                evaluating_ref.value = False

        return cast(ComputedResultProtocol[_T], AsyncComputedResult(current, fn))

    return wrap_cp


class AsyncComputedResult(Generic[_T], ReadableMixin[_T]):
    __slot__ = ("_result", "_fn")

    def __init__(self, result: TSignal[_T], fn: _T_async_fn[_T]) -> None:
        self._result = result
        self._fn = fn

    @property
    def value(self):
        return self._result.value

    def __call__(self) -> Any:
        return self._fn()



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    TypeVar,
    Callable,
    Optional,
    Generic,
    Union,
    cast,
    overload,
)
from signe.core.consts import EffectState
from signe.core.context import get_default_scheduler
from signe.core.deps import GetterDepManager
from signe.core.helper import has_changed
from signe.core.idGenerator import IdGen

from signe.core.protocols import ComputedResultProtocol
from signe.core.mixins import ReadableMixin

from .effect import Effect
from .scope import Scope, ScopeSuite, _DEFAULT_SCOPE_SUITE

if TYPE_CHECKING:  # pragma: no cover
    from .runtime import ExecutionScheduler

_T = TypeVar("_T")


class Computed(Generic[_T], ReadableMixin[_T]):
    _id_gen = IdGen("Computed")

    def __init__(
        self,
        fn: Callable[[], _T],
        *,
        scheduler: ExecutionScheduler,
        scope: Union[Scope, ScopeSuite],
        debug_trigger: Optional[Callable] = None,
        priority_level=1,
        debug_name: Optional[str] = None,
        capture_parent_effect=True,
    ) -> None:
        self.__id = self._id_gen.new()
        self._value = None
        self._debug_name = debug_name

        def getter():
            return fn()

        def trigger_fn(_):
            self.trigger(EffectState.PENDING)

        self._effect = Effect(
            getter,
            scheduler=scheduler,
            trigger_fn=trigger_fn,
            # immediate=False,
            debug_trigger=debug_trigger,
            scope=scope,
            state=EffectState.COMPUTED_INIT,
            debug_name=debug_name,
            capture_parent_effect=capture_parent_effect,
        )
        self._dep_manager = GetterDepManager(scheduler)

        if isinstance(scope, Scope):
            scope.add_disposable(self)
        elif isinstance(scope, ScopeSuite):
            scope.mark_with_scope(self)

    @property
    def id(self):
        return self.__id  # pragma: no cover

    def trigger(self, state: EffectState):
        state = EffectState.PENDING if state == EffectState.NEED_UPDATE else state
        self._effect.update_state(state)  # type: ignore

        self._dep_manager.triggered("value", self._value, state)

    def confirm_state(self):
        if self._effect.state <= EffectState.NEED_UPDATE:  # type: ignore
            self._update_value()

    def dispose(self):
        self._effect = None
        self._value = None
        self._dep_manager.dispose()

    @property
    def value(self):
        if self._effect.state <= EffectState.NEED_UPDATE:  # type: ignore
            self._update_value()

        self._dep_manager.tracked("value", computed=self)
        return self._value

    def __call__(self) -> _T:
        return self.value  # type: ignore

    def _update_value(self):
        new_value = self._effect.update()  # type: ignore

        if has_changed(self._value, new_value):
            self._dep_manager.triggered("value", new_value, EffectState.NEED_UPDATE)

        self._value = new_value

    def __repr__(self) -> str:
        state = self._effect.state  # type: ignore
        return f"Computed(id ={self.id}, name={self._debug_name}),state={state}"


_T_computed = ComputedResultProtocol[_T]
_T_computed_setter = Callable[[Callable[[], _T]], _T_computed]


@overload
def computed(
    fn: None = ...,
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> _T_computed_setter:
    ...


@overload
def computed(
    fn: Callable[[], _T],
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> _T_computed[_T]:
    ...


def computed(
    fn: Optional[Callable[[], _T]] = None,
    *,
    priority_level=1,
    debug_trigger: Optional[Callable] = None,
    debug_name: Optional[str] = None,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
) -> Union[_T_computed_setter, _T_computed[_T]]:
    kws = {
        "priority_level": priority_level,
        "debug_trigger": debug_trigger,
        "debug_name": debug_name,
        "scheduler": scheduler or get_default_scheduler(),
    }

    if fn:
        cp = Computed(fn, **kws, scope=scope or _DEFAULT_SCOPE_SUITE)
        return cast(ComputedResultProtocol[_T], cp)
    else:

        def wrap_cp(fn: Callable[[], _T]):
            return computed(fn, **kws, scope=scope)

        return wrap_cp



from itertools import count


class IdGen:
    def __init__(self, name: str):
        self._name = name
        self._num = count()

    def new(self):
        return f"{self._name}_{next(self._num)}"



import asyncio
from dataclasses import dataclass
from signe.core.context import get_default_scheduler
from signe.core.effect import Effect
from signe.core.helper import has_changed, get_func_args_count
from signe.core.reactive import is_reactive, track_all
from signe.core.scope import Scope, _DEFAULT_SCOPE_SUITE, ScopeSuite
from typing import (
    Any,
    Dict,
    List,
    Sequence,
    TypeVar,
    Callable,
    Union,
    cast,
    overload,
    Optional,
    Generic,
)

from .types import TGetter, TSignal
from .mixins import is_signal

from .runtime import ExecutionScheduler


T = TypeVar("T")


class OnGetterModel(Generic[T]):
    __slots__ = ("_ref", "_fn")

    def __init__(
        self,
        ref: Union[TSignal, Callable[[], T]],
        scheduler: ExecutionScheduler,
        deep=False,
    ) -> None:
        self._ref = ref

        # with track
        if is_signal(ref):

            def getter():
                value = self._ref.value
                if is_reactive(value):
                    track_all(value, scheduler, deep)
                return value

            self._fn = getter
        else:

            def getter_reactive_fn():
                obj = cast(Callable, ref)()
                track_all(obj, scheduler, deep)
                return obj

            self._fn = getter_reactive_fn

    def get_value(self):
        return self._fn()


@dataclass(frozen=True)
class WatchedState:
    __slot__ = ("current", "previous")
    current: Any
    previous: Any


@overload
def on(
    source: Union[TGetter, Sequence[TGetter]],
    fn: Optional[Callable[..., None]] = None,
    *,
    onchanges=False,
    effect_kws: Optional[Dict[str, Any]] = None,
    deep=False,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
):
    ...


@overload
def on(
    source: Union[TGetter, Sequence[TGetter]],
    fn: Optional[Callable[..., None]] = None,
    *,
    onchanges=False,
    deep=False,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
):
    ...


def on(
    source: Union[TGetter, Sequence[TGetter]],
    fn: Optional[Callable[..., None]] = None,
    *,
    onchanges=False,
    effect_kws: Optional[Dict[str, Any]] = None,
    deep=False,
    scope: Optional[Union[Scope, ScopeSuite]] = None,
    scheduler: Optional[ExecutionScheduler] = None,
):
    call_kws = {
        "onchanges": onchanges,
        "effect_kws": effect_kws,
        "deep": deep,
        "scheduler": scheduler or get_default_scheduler(),
    }

    if fn is None:

        def wrap_cp(fn: Callable[[], T]):
            return on(source, fn, **call_kws, scope=scope)  # type: ignore

        return wrap_cp

    if asyncio.iscoroutinefunction(fn):
        base_fn = fn

        def wrap_fn(*args, **kws):  # type: ignore
            asyncio.get_event_loop().create_task(base_fn(*args, **kws))  # type: ignore

        fn = wrap_fn

    getters: List[OnGetterModel] = []
    if isinstance(source, Sequence):
        getters = [OnGetterModel(g, call_kws["scheduler"], deep) for g in source]  # type: ignore
    else:
        getters = [OnGetterModel(source, call_kws["scheduler"], deep)]  # type: ignore

    def getter():
        return [g.get_value() for g in getters]

    args_count = get_func_args_count(fn)
    prev_values = [None] * len(getters)

    def scheduler_fn(effect: Effect):
        nonlocal prev_values
        if (not effect._active) or (not effect.is_need_update()):
            return

        new_values = effect.update()
        if deep or any(has_changed(n, v) for n, v in zip(new_values, prev_values)):
            states = (
                WatchedState(cur, prev)
                for cur, prev in zip(new_values, prev_values or new_values)
            )
            if args_count == 0:
                fn()
            else:
                fn(*states)

        prev_values = new_values

    effect = Effect(
        getter,
        scheduler=scheduler or get_default_scheduler(),
        scheduler_fn=scheduler_fn,
        **(effect_kws or {}),
        scope=scope or _DEFAULT_SCOPE_SUITE,
    )

    if onchanges:
        prev_values = effect.update()
    else:
        scheduler_fn(effect)



from __future__ import annotations
from typing import (
    Callable,
    Generic,
    Protocol,
    TypeVar,
    TYPE_CHECKING,
    runtime_checkable,
)


if TYPE_CHECKING:
    from signe.core.deps import Dep
    from .consts import EffectState

_TOut = TypeVar("_TOut", covariant=True)
_T = TypeVar("_T")


@runtime_checkable
class OnGetterProtocol(Protocol):
    def get_value_with_track(self):
        ...

    def get_value_without_track(self):
        ...


class DisposableProtocol(Protocol):
    def dispose(self):
        ...


class GetterProtocol(Protocol[_TOut]):  # type: ignore
    @property
    def id(self) -> str:
        ...

    @property
    def is_signal(self) -> bool:
        ...

    @property
    def value(self) -> _TOut:
        ...

    def confirm_state(self):
        ...


class CallerProtocol(Protocol[_TOut]):  # type: ignore
    @property
    def id(self) -> str:
        ...

    def trigger(self, state: EffectState):
        ...

    @property
    def is_effect(self) -> bool:
        ...

    @property
    def state(self) -> EffectState:
        ...

    def update(self) -> _TOut:
        ...

    def add_upstream_ref(self, dep: Dep):
        ...

    def add_cleanup(self, fn: Callable[[], None]):
        ...


# class IScope(Protocol):
#     def add_disposable(self, disposable: DisposableProtocol):
#         ...

#     def dispose(self):
#         ...


class SignalResultProtocol(Protocol[_T]):
    @property
    def value(self) -> _T:
        ...

    @value.setter
    def value(self, value: _T):
        ...

    def set_value(self, value: _T):
        ...


class ComputedResultProtocol(Generic[_TOut], Protocol):  # type: ignore
    @property
    def value(self) -> _TOut:
        ...

    def __call__(
        self,
    ) -> _TOut:
        ...


@runtime_checkable
class RawableProtocol(Protocol[_TOut]):
    def to_raw(self) -> _TOut:
        ...
